@@ -2,7 +2,7 @@ from __future__ import annotations
 
 from typing import cast
 
-__version__ = "2.5.0"
+__version__ = "2.8.4"
 
 
 import asyncio
@@ -17,7 +17,7 @@ import async_timeout
 from bleak import BleakClient
 from bleak.backends.device import BLEDevice
 from bleak.backends.service import BleakGATTServiceCollection
-from bleak.exc import BleakDBusError, BleakError
+from bleak.exc import BleakDBusError, BleakDeviceNotFoundError, BleakError
 
 DISCONNECT_TIMEOUT = 5
 
@@ -36,14 +36,16 @@ if IS_LINUX:
 NO_RSSI_VALUE = -127
 
 
-# Make sure bleak and dbus-next have time
+# Make sure bleak and dbus-fast have time
 # to run their cleanup callbacks or the
 # retry call will just fail in the same way.
+BLEAK_TRANSIENT_BACKOFF_TIME = 0.25
+BLEAK_TRANSIENT_MEDIUM_BACKOFF_TIME = 0.55
+BLEAK_TRANSIENT_LONG_BACKOFF_TIME = 1.25
 BLEAK_DBUS_BACKOFF_TIME = 0.25
-BLEAK_OUT_OF_SLOTS_BACKOFF_TIME = 1.5
+BLEAK_OUT_OF_SLOTS_BACKOFF_TIME = 4.00
 BLEAK_BACKOFF_TIME = 0.1
 
-
 RSSI_SWITCH_THRESHOLD = 5
 
 __all__ = [
@@ -79,24 +81,41 @@ MAX_TRANSIENT_ERRORS = 9
 # to finish in < 60s or declare we cannot connect
 
 MAX_CONNECT_ATTEMPTS = 4
-BLEAK_TIMEOUT = 16.25
+BLEAK_TIMEOUT = 20.0
 
 # Bleak may not always timeout
 # since the dbus connection can stall
 # so we have an additional timeout to
 # be sure we do not block forever
 # This is likely fixed in https://github.com/hbldh/bleak/pull/1092
-BLEAK_SAFETY_TIMEOUT = 20
+BLEAK_SAFETY_TIMEOUT = 21.0
 
-# These errors are transient with dbus, and we should retry
-TRANSIENT_ERRORS = {"le-connection-abort-by-local", "br-connection-canceled"}
+TRANSIENT_ERRORS_LONG_BACKOFF = {
+    "ESP_GATT_ERROR",
+}
+
+TRANSIENT_ERRORS_MEDIUM_BACKOFF = {
+    "ESP_GATT_CONN_FAIL_ESTABLISH",
+}
 
 DEVICE_MISSING_ERRORS = {"org.freedesktop.DBus.Error.UnknownObject"}
 
 OUT_OF_SLOTS_ERRORS = {"available connection", "connection slot"}
 
+TRANSIENT_ERRORS = {
+    "le-connection-abort-by-local",
+    "br-connection-canceled",
+    "ESP_GATT_CONN_FAIL_ESTABLISH",
+    "ESP_GATT_CONN_TERMINATE_PEER_USER",
+    "ESP_GATT_CONN_TERMINATE_LOCAL_HOST",
+    "ESP_GATT_CONN_CONN_CANCEL",
+} | OUT_OF_SLOTS_ERRORS
+
 # Currently the same as transient error
-ABORT_ERRORS = TRANSIENT_ERRORS
+ABORT_ERRORS = (
+    TRANSIENT_ERRORS | TRANSIENT_ERRORS_MEDIUM_BACKOFF | TRANSIENT_ERRORS_LONG_BACKOFF
+)
+
 
 ABORT_ADVICE = (
     "Interference/range; "
@@ -158,9 +177,15 @@ def ble_device_has_changed(original: BLEDevice, new: BLEDevice) -> bool:
 
 def ble_device_description(device: BLEDevice) -> str:
     """Get the device description."""
-    if isinstance(device.details, dict) and "path" in device.details:
-        return device.details["path"]
-    return device.address
+    details = device.details
+    address = device.address
+    if isinstance(details, dict):
+        if path := details.get("path"):
+            # /org/bluez/hci2
+            return f"{address} -> {path[0:15]}"
+        if source := details.get("source"):
+            return f"{address} -> {source}"
+    return address
 
 
 def _get_possible_paths(path: str) -> Generator[str, None, None]:
@@ -196,14 +221,28 @@ def address_to_bluez_path(address: str, adapter: str | None = None) -> str:
 
 def calculate_backoff_time(exc: Exception) -> float:
     """Calculate the backoff time based on the exception."""
+
     if isinstance(
         exc, (BleakDBusError, EOFError, asyncio.TimeoutError, BrokenPipeError)
     ):
         return BLEAK_DBUS_BACKOFF_TIME
-    if isinstance(exc, BleakError) and any(
-        error in str(exc) for error in OUT_OF_SLOTS_ERRORS
-    ):
+    # If the adapter runs out of slots can get a BleakDeviceNotFoundError
+    # since the device is no longer visible on the adapter. Almost none of
+    # the adapters document how many connection slots they have so we cannot
+    # know if we are out of slots or not. We can only guess based on the
+    # error message and backoff.
+    if isinstance(exc, BleakDeviceNotFoundError):
         return BLEAK_OUT_OF_SLOTS_BACKOFF_TIME
+    if isinstance(exc, BleakError):
+        bleak_error = str(exc)
+        if any(error in bleak_error for error in OUT_OF_SLOTS_ERRORS):
+            return BLEAK_OUT_OF_SLOTS_BACKOFF_TIME
+        if any(error in bleak_error for error in TRANSIENT_ERRORS_MEDIUM_BACKOFF):
+            return BLEAK_TRANSIENT_MEDIUM_BACKOFF_TIME
+        if any(error in bleak_error for error in TRANSIENT_ERRORS_LONG_BACKOFF):
+            return BLEAK_TRANSIENT_LONG_BACKOFF_TIME
+        if any(error in bleak_error for error in TRANSIENT_ERRORS):
+            return BLEAK_TRANSIENT_BACKOFF_TIME
     return BLEAK_BACKOFF_TIME
 
 
@@ -402,9 +441,22 @@ async def wait_for_disconnect(device: BLEDevice, min_wait_time: float) -> None:
         )
         if min_wait_time and waited < min_wait_time:
             await asyncio.sleep(min_wait_time - waited)
-    except KeyError:
+    except KeyError as ex:
         # Device was removed from bus
-        pass
+        #
+        # In testing it was found that most of the CSR adapters
+        # only support 5 slots and the broadcom only support 7 slots.
+        #
+        # When they run out of slots the device they are trying to
+        # connect to disappears from the bus so we must backoff
+        _LOGGER.debug(
+            "%s - %s: Device was removed from bus, waiting %s for it to re-appear: %s",
+            device.name,
+            ble_device_description(device),
+            min_wait_time,
+            ex,
+        )
+        await asyncio.sleep(min_wait_time)
     except Exception:  # pylint: disable=broad-except
         _LOGGER.debug(
             "%s - %s: Failed waiting for disconnect",
@@ -450,14 +502,14 @@ async def establish_connection(
         if isinstance(exc, asyncio.TimeoutError) or "not found" in str(exc):
             raise BleakNotFoundError(msg) from exc
         if isinstance(exc, BleakError):
-            if any(error in str(exc) for error in ABORT_ERRORS):
-                raise BleakAbortedError(f"{msg}: {ABORT_ADVICE}") from exc
-            if any(error in str(exc) for error in DEVICE_MISSING_ERRORS):
-                raise BleakNotFoundError(f"{msg}: {DEVICE_MISSING_ADVICE}") from exc
             if any(error in str(exc) for error in OUT_OF_SLOTS_ERRORS):
                 raise BleakOutOfConnectionSlotsError(
                     f"{msg}: {OUT_OF_SLOTS_ADVICE}"
                 ) from exc
+            if any(error in str(exc) for error in ABORT_ERRORS):
+                raise BleakAbortedError(f"{msg}: {ABORT_ADVICE}") from exc
+            if any(error in str(exc) for error in DEVICE_MISSING_ERRORS):
+                raise BleakNotFoundError(f"{msg}: {DEVICE_MISSING_ADVICE}") from exc
         raise BleakConnectionError(msg) from exc
 
     create_client = True
@@ -564,7 +616,11 @@ async def establish_connection(
             _raise_if_needed(name, description, exc)
         except BLEAK_EXCEPTIONS as exc:
             bleak_error = str(exc)
-            if any(error in bleak_error for error in TRANSIENT_ERRORS):
+            # BleakDeviceNotFoundError can mean that the adapter has run out of
+            # connection slots.
+            if isinstance(exc, BleakDeviceNotFoundError) or any(
+                error in bleak_error for error in TRANSIENT_ERRORS
+            ):
                 transient_errors += 1
             else:
                 connect_errors += 1
